/*****************************************************************************

  FileName:	 Q931.c

  Contents:	 Implementation of Q.931 stack main interface functions. 
				See q931.h for description. 

  License/Copyright:

  Copyright (c) 2007, Jan Vidar Berger, Case Labs, Ltd. All rights reserved.
  email:janvb@caselaboratories.com  

  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions are 
  met:

	* Redistributions of source code must retain the above copyright notice, 
	  this list of conditions and the following disclaimer.
	* Redistributions in binary form must reproduce the above copyright notice, 
	  this list of conditions and the following disclaimer in the documentation 
	  and/or other materials provided with the distribution.
	* Neither the name of the Case Labs, Ltd nor the names of its contributors 
	  may be used to endorse or promote products derived from this software 
	  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
  POSSIBILITY OF SUCH DAMAGE.
*****************************************************************************/

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#define Q931PRIVATE

#include "Q921.h"
#include "Q931.h"

#if defined(HAVE_DIALECT_5ESS)
#include "5ESS.h"
#endif

#if defined(HAVE_DIALECT_DSS1)
#include "DSS1.h"
#endif

/*****************************************************************************

  Dialect function pointers tables.

  The following function pointer arrays define pack/unpack functions and 
  processing furnctions for the different Q.931 based dialects.

  The arrays are initialized with pointers to dummy functions and later
  overrided with pointers to actual functions as new dialects are added.

  The initial Q.931 will as an example define 2 dialects as it treats User
  and Network mode as separate ISDN dialects.

  The API messages Q931AddProc, Q931AddMes, Q931AddIE are used to initialize
  these table entries during system inititialization of a stack.

*****************************************************************************/

/*****************************************************************************

  Core system tables and variables.

*****************************************************************************/

L3INT Q931L4HeaderSpace = {0};	/* header space to be ignoder/inserted */
				/* at head of each message. */

L3INT Q931L2HeaderSpace = {4};	/* Q921 header space, sapi, tei etc */

/*****************************************************************************

  Main interface callback functions. 

*****************************************************************************/

L3ULONG (*Q931GetTimeProc) (void) = NULL;	/* callback for func reading time in ms */

/*****************************************************************************

  Function:	 Q931SetL4HeaderSpace

  Description:  Set the # of bytes to be inserted/ignored at the head of
		each message. Q931 will issue a message with space for header
		and the user will use this to fill in whatever header info
		is required to support the architecture used.

*****************************************************************************/
void Q931SetL4HeaderSpace(L3INT space)
{
	Q931L4HeaderSpace = space;
}

/*****************************************************************************

  Function:	 Q931SetL2HeaderSpace

  Description:  Set the # of bytes to be inserted/ignored at the head of
		each message. Q931 will issue a message with space for header
		and the user will use this to fill in whatever header info
		is required to support the architecture used.

*****************************************************************************/
void Q931SetL2HeaderSpace(L3INT space)
{
	Q931L2HeaderSpace = space;
}

/*****************************************************************************

  Function:	 Q931ProcDummy

  Description:  Dummy function for message processing.

*****************************************************************************/
L3INT Q931ProcDummy(Q931_TrunkInfo_t *pTrunk, L3UCHAR * b,L3INT c)
{
	return Q931E_INTERNAL;
}

/*****************************************************************************

  Function:	 Q931UmesDummy

  Description:  Dummy function for message processing

*****************************************************************************/
L3INT Q931UmesDummy(Q931_TrunkInfo_t *pTrunk,L3UCHAR *IBuf, Q931mes_Generic *OBuf, L3INT IOff, L3INT Size)
{
	return Q931E_UNKNOWN_MESSAGE;
}

/*****************************************************************************

  Function:	 Q931UieDummy

  Description:  Dummy function for message processing

*****************************************************************************/
L3INT Q931UieDummy(Q931_TrunkInfo_t *pTrunk, Q931mes_Generic *pMsg, L3UCHAR * IBuf, L3UCHAR * OBuf, L3INT *IOff, L3INT *OOff)
{
	return Q931E_UNKNOWN_IE;
}

/*****************************************************************************

  Function:	 Q931PmesDummy

  Description:  Dummy function for message processing

*****************************************************************************/
L3INT Q931PmesDummy(Q931_TrunkInfo_t *pTrunk, Q931mes_Generic *IBuf, L3INT ISize, L3UCHAR *OBuf, L3INT *OSize)
{
	return Q931E_UNKNOWN_MESSAGE;
}

/*****************************************************************************

  Function:	 Q931PieDummy

  Description:  Dummy function for message processing

*****************************************************************************/
L3INT Q931PieDummy(Q931_TrunkInfo_t *pTrunk,L3UCHAR *IBuf, L3UCHAR *OBuf, L3INT *Octet)
{
	return Q931E_UNKNOWN_IE;
}

/*****************************************************************************

  Function:	 Q931TxDummy

  Description:  Dummy function for message processing

*****************************************************************************/
L3INT Q931TxDummy(Q931_TrunkInfo_t *pTrunk, L3UCHAR * b, L3INT n)
{
	return Q931E_MISSING_CB;
}

/*****************************************************************************

  Function:	 Q931Initialize

  Description:  This function Initialize the stack. 
  
		Will set up the trunk array, channel
		arrays and initialize Q931 function arrays before it finally
		set up EuroISDN processing with User as dialect 0 and 
		Network as dialect 1.

  Note:		 Initialization of other stacks should be inserted after
		the initialization of EuroISDN.

*****************************************************************************/
/**
 * Q931Initialize
 * \brief	Initialize Q.931 stack
 */
Q931_API(void) Q931Initialize(void)
{
	/*
	 * Initialize Q.931 dialect registry and register dialects
	 */
	Q931DialectRegistryInit();

	/* Register Q.931 */
	Q931DialectRegister(Q931_Dialect_Q931, Q931_TE, Q931CreateTE);
	Q931DialectRegister(Q931_Dialect_Q931, Q931_NT, Q931CreateNT);

	/* Register DSS1 */
	Q931DialectRegister(Q931_Dialect_DSS1, Q931_TE, DSS1CreateTE);
	Q931DialectRegister(Q931_Dialect_DSS1, Q931_NT, DSS1CreateNT);

#if 0
	/* Register 5ESS */
	Q931DialectRegister(Q931_Dialect_5ESS, Q931_TE, ATT5ESSCreateTE);
	Q931DialectRegister(Q931_Dialect_5ESS, Q931_NT, ATT5ESSCreateNT);
#endif
}

/**
 * Q931TimerTick
 * \brief	Periodically called to update and check for expired timers
 * \param	pTrunk	Q.931 trunk
 */
Q931_API(void) Q931TimerTick(Q931_TrunkInfo_t *pTrunk)
{
	struct Q931_Call *call = NULL;
	L3ULONG now = 0;
	L3INT x;

	/* TODO: Loop through all active calls, check timers and call timout procs
	 * if timers are expired.
	 * Implement a function array so each dialect can deal with their own
	 * timeouts.
	 */
	now = Q931GetTime();

	for (x = 0; x < Q931MAXCALLPERTRUNK; x++) {
		call = &pTrunk->call[x];

		if (!call->InUse || !call->Timer || !call->TimerID)
			continue;

		if (call->Timer <= now) {
			int id = call->TimerID;	/* Q931CallStopTimer resets the TimerID */

			/* Stop Timer */
			Q931CallStopTimer(call, id);

			/* Increment expire counter */
			Q931CallIncrementTimerExpireCount(call);

			/* Invoke dialect timeout callback */
			Q931Timeout(pTrunk, call, id);

			/*
			 * Release CRV if call has gone back to Null state after timer event
			 * and send an event to L4 before doing so
			 */
			if (!Q931CallIsGlobal(call)) {
				if ((Q931CallGetState(call) & 0xff) == Q931_U0) {
					/* Send event, so L4 can clean up */
					struct Q931_CallEvent event;

					event.id    = Q931_EVENT_RELEASE_CRV;
					event.type  = Q931_EVENT_TYPE_CRV;
					event.error = 0;

					Q931CallSendEvent(call, &event);

					Q931Log(pTrunk, Q931_LOG_DEBUG, "Garbage-collecting call %d\n", Q931CallGetCRV(call));

					/* Release call */
					Q931CallRelease(call);
				}
			}
		}
	}
}

/*****************************************************************************

  Function:	 Q931Rx23

  Description:  Receive message from layer 2 (LAPD). Receiving a message 
				is always done in 2 steps. First the message must be 
				interpreted and translated to a static struct. Secondly
				the message is processed and responded to.

				The Q.931 message contains a static header that is 
				interpreted in this function. The rest is interpreted
				in a sub function according to mestype.

  Parameters:   pTrunk  [IN]	Ptr to trunk info.
				buf	 [IN]	Ptr to buffer containing message.
				Size	[IN]	Size of message.

  Return Value: Error Code. 0 = No Error, < 0 :error, > 0 : Warning
				see q931errors.h for details.

*****************************************************************************/
Q931_API(L3INT) Q931Rx23(Q931_TrunkInfo_t *pTrunk, L3INT ind, L3UCHAR tei, L3UCHAR * buf, L3INT Size)
{
	L3UCHAR *Mes = NULL;
	L3INT RetCode = Q931E_NO_ERROR;
	Q931mes_Generic *m = (Q931mes_Generic *) pTrunk->L3Buf;
	struct Q931_Call *call = NULL;
	L3INT ISize;
	L3INT IOff = 0;
	L3INT L2HSize = Q931L2HeaderSpace;

	switch (ind) {
	case Q921_DL_UNIT_DATA:		/* DL-UNITDATA indication (UI frame, 3 byte header) */
		L2HSize = 3;

	case Q921_DL_DATA:		/* DL-DATA indication (I frame, 4 byte header) */
		/* Reset our decode buffer */
		memset(pTrunk->L3Buf, 0, sizeof(pTrunk->L3Buf));

		/* L2 Header Offset */
		Mes = &buf[L2HSize];

		/* Protocol Discriminator */
		m->ProtDisc = Mes[IOff++];

		/* CRV */
		m->CRVFlag = (Mes[IOff + 1] >> 7) & 0x01;
		m->CRV = Q931Uie_CRV(pTrunk, Mes, m->buf, &IOff, &ISize);

		/* Message Type */
		m->MesType = Mes[IOff++];

		/* Store tei */
		m->Tei = tei;

		/* d'oh a little ugly but this saves us from:
		 *	a) doing Q.921 work in the lower levels (extracting the TEI ourselves)
		 *	b) adding a tei parameter to _all_ Proc functions
		 * TODO: a little ugly, merge into next codeblock
		 */
		if (tei && m->CRV) {
			/* Find the call using CRV */
			call = Q931GetCallByCRV(pTrunk, m->CRV);
			if (call && !call->Tei) {
				call->Tei = tei;
			}
		}

		/* vvv new code start here */
		if (m->CRV) {
			/* Lookup call reference */
			call = Q931GetCallByCRV(pTrunk, m->CRV);
			if (!call) {
				/* oops, no call found */
				if(m->MesType == Q931mes_SETUP) {
					/* Setup message, create new call handle */
					call = Q931CallNewIncoming(pTrunk, m->CRV);
				} else {
					/* Not a setup message -> INVALID CRV */
					/* TODO: respond with proper RELEASE COMPLETE + cause combination */
					return Q931E_INVALID_CRV;
				}

				/* Still no call handle? */
				if (!call) {
					/* blah, failed to allocate new call -> Disconnect */
					Q931Disconnect(pTrunk, MSGF_FROM_L2, m->CRV, 42);
					return Q931E_INTERNAL;
				}
			}

			/* Diagnostic message */
			Q931Log(pTrunk, Q931_LOG_DEBUG, "Call is in state %d [%s]\n", Q931CallGetState(call), Q931CallGetStateName(call));
		} else {
			/* Global CRV ?? */
			call = Q931GetGlobalCall(pTrunk);

			Q931Log(pTrunk, Q931_LOG_DEBUG, "Using global call for message\n");
		}
		/* ^^^ new code ends here */

		Q931Log(pTrunk, Q931_LOG_DEBUG, "Received message from Q.921 (ind %d, tei %d, size %d)\nMesType: %d, CRVFlag %d (%s), CRV %d (Dialect: %s)\n", ind, m->Tei, Size,
						 m->MesType, m->CRVFlag, m->CRVFlag ? "Terminator" : "Originator", m->CRV, Q931DialectGetName(pTrunk->Dialect));

		RetCode = Q931Umes(pTrunk, m->MesType, Mes, (Q931mes_Generic *)pTrunk->L3Buf, IOff, Size - L2HSize);
		if (RetCode >= Q931E_NO_ERROR) {
			RetCode = Q931Proc(pTrunk, pTrunk->L3Buf, 2);
			Q931Log(pTrunk, Q931_LOG_DEBUG, "Q931Proc[%s][%d]() returned %d\n", Q931DialectGetName(pTrunk->Dialect), m->MesType, RetCode);
		} else {
			Q931Log(pTrunk, Q931_LOG_DEBUG, "Q931Umes[%s][%d]() returned %d\n", Q931DialectGetName(pTrunk->Dialect), m->MesType, RetCode);
		}

		/* Error handling for Umes() and Proc() */
		if (RetCode < Q931E_NO_ERROR) {
			/* Initialize response message */

			/* TODO: Add Error codes for Mandatory IE missing and error */
			switch (RetCode) {
			case Q931E_MANDATORY_IE_MISSING:
				/* Send STATUS with cause 96 */
				RetCode = Q931StatusEnquiryResponse(pTrunk, buf, call, Q850_CAUSE_MANDATORY_IE_MISSING);
				break;

			case Q931E_ILLEGAL_IE:			/* ?? */
				/* Send STATUS with cause 96 */
				RetCode = Q931StatusEnquiryResponse(pTrunk, buf, call, Q850_CAUSE_IE_NONEXIST);
				break;

			case Q931E_UNKNOWN_IE:
				/* Send STATUS with cause 100 */
				RetCode = Q931StatusEnquiryResponse(pTrunk, buf, call, Q850_CAUSE_INVALID_IE_CONTENTS);
				break;

			case Q931E_UNEXPECTED_MESSAGE:
			case Q931E_UNRECOGNIZED_MESSAGE:
				/* Ignore in State 0 */
				if ((Q931CallGetState(call) & 0xff) == Q931_U0)
					goto out;

				/* Ignore for RELEASE and RELEASE COMPLETE */
				if (m->MesType == Q931mes_RELEASE || m->MesType == Q931mes_RELEASE_COMPLETE)
					goto out;

				if (Q931TrunkGetStatusEnquiry(pTrunk)) {
					/* Send STATUS ENQUIRY */
					Q931InitMesGeneric(m);
					m->CRV     = Q931CallGetCRV(call);
					m->CRVFlag = Q931CallIsOutgoing(call) ? 0 : 1;
					m->MesType = Q931mes_STATUS_ENQUIRY;

					/* TODO: check and / or fire up T322 */
					RetCode = Q931Tx32Data(pTrunk, 0, buf, m->Size);
				} else {
					/* Send STATUS with cause X */
					RetCode = Q931StatusEnquiryResponse(pTrunk, buf, call, 
											(RetCode == Q931E_UNEXPECTED_MESSAGE) ? Q850_CAUSE_WRONG_MESSAGE : Q850_CAUSE_MESSAGE_TYPE_NONEXIST);
				}
				break;

			default:	/* unhandled / unknown error (ignore) */
				break;
			}
		}
		break;

	default:
		break;
	}

out:
	if (call) {
		/* Send queued call events to layer 4 (if any) */
		Q931CallSendQueuedEvents(call);

		/*
		 * Release CRV if call is still (or again) in Null state,
		 * and send an event to L4 before doing so
		 *
		 * TODO: Remove Q931CallRelease() calls from Q931Proc_*
		 *       functions
		 */
		if (!Q931CallIsGlobal(call)) {
			if ((Q931CallGetState(call) & 0xff) == Q931_U0) {
				/* Send event, so L4 can clean up */
				struct Q931_CallEvent event;

				event.id    = Q931_EVENT_RELEASE_CRV;
				event.type  = Q931_EVENT_TYPE_CRV;
				event.error = 0;

				Q931CallSendEvent(call, &event);

				Q931Log(pTrunk, Q931_LOG_DEBUG, "Garbage-collecting call %d\n", Q931CallGetCRV(call));

				/* Release call */
				Q931CallRelease(call);
			}
		}
	}
	return RetCode;
}

/*****************************************************************************

  Function:	 Q931Tx34

  Description:  Called from the stack to send a message to layer 4.

  Parameters:   Mes[IN]	 Ptr to message buffer.
		Size[IN]	Message size in bytes.

  Return Value: Error Code. 0 = No Error, < 0 :error, > 0 : Warning
		see q931errors.h for details.

*****************************************************************************/
Q931_API(L3INT) Q931Tx34(Q931_TrunkInfo_t *pTrunk, L3UCHAR * Mes, L3INT Size)
{
	Q931Log(pTrunk, Q931_LOG_DEBUG, "Sending message to Layer4 (size: %d)\n", Size);

	if (pTrunk->Q931Tx34CBProc) {
		return pTrunk->Q931Tx34CBProc(pTrunk->PrivateData34, Mes, Size);
	}
	return Q931E_MISSING_CB;	
}

/*****************************************************************************

  Function:	 Q931Rx43

  Description:  Receive message from Layer 4 (application).

  Parameters:   pTrunk[IN]  Trunk #.
		buf[IN]	 Message Pointer.
		Size[IN]	Message size in bytes.

  Return Value: Error Code. 0 = No Error, < 0 :error, > 0 : Warning
		see q931errors.h for details.

*****************************************************************************/
Q931_API(L3INT) Q931Rx43(Q931_TrunkInfo_t *pTrunk,L3UCHAR * buf, L3INT Size)
{
	Q931mes_Header *ptr = (Q931mes_Header*)&buf[Q931L4HeaderSpace];
	L3INT RetCode = Q931E_NO_ERROR;
	struct Q931_Call *call = NULL;

	Q931Log(pTrunk, Q931_LOG_DEBUG, "Receiving message from Layer4 (size: %d, type: %d [0x%x], CRV: %d, CRVFlag: %d)\n", Size, ptr->MesType, ptr->MesType, ptr->CRV, ptr->CRVFlag);

	/*
	 * Check if message is valid in current state & dialect
	 */

	/* Get call handle */
	if (ptr->CRV) {
		/* Find the call using CRV */
		call = Q931GetCallByCRV(pTrunk, ptr->CRV);
		if(!call)
			return Q931E_INVALID_CRV;

		/* SAFEGUARD: make sure the CRVFlag is set correctly */
		ptr->CRVFlag = Q931CallIsOutgoing(call) ? 0 : 1;
	} else {
		/* global call handle */
		call = Q931GetGlobalCall(pTrunk);
	}

	/* Check if message can be handled */
	if (!Q931DialectGetMesProc(pTrunk->Dialect, ptr->MesType)) {
		RetCode = Q931E_UNRECOGNIZED_MESSAGE;

		Q931Log(pTrunk, Q931_LOG_WARNING, "Unrecognized message from Layer 4 (size: %d, type: %d)\n", Size, ptr->MesType);

	        /* Ignore unknown message? */
	        if (Q931TrunkIsSetFlag(pTrunk, Q931_TFLAG_IGNORE_UNKNOWN_MSG)) {
			RetCode = Q931E_NO_ERROR;
		}
		goto out;
	}

	/* Check if message is valid in current call state */
	if (!Q931DialectIsEventLegal(pTrunk->Dialect, Q931CallGetState(call), ptr->MesType, 4)) {
		RetCode = Q931E_UNEXPECTED_MESSAGE;

		Q931Log(pTrunk, Q931_LOG_ERROR, "Unexpected message (invalid in current state) from Layer 4 (size: %d, type: %d [0x%x])\n", Size, ptr->MesType, ptr->MesType);
		goto out;
	}

	Q931Log(pTrunk, Q931_LOG_NOTICE, "Handling message from Layer4 in %s call state: %d\n", Q931CallIsGlobal(call) ? "global" : "", Q931CallGetState(call));

	RetCode = Q931Proc(pTrunk, buf, 4);

out:
	if (call) {
		/*
		 * Release CRV if call is still (or again) in Null state,
		 * and send an event to L4 before doing so
		 *
		 * TODO: Remove Q931CallRelease() calls from Q931Proc_*
		 *       functions
		 */
		if (!Q931CallIsGlobal(call)) {
			if ((Q931CallGetState(call) & 0xff) == Q931_U0) {
				/* Send event, so L4 can clean up */
				struct Q931_CallEvent event;

				event.id    = Q931_EVENT_RELEASE_CRV;
				event.type  = Q931_EVENT_TYPE_CRV;
				event.error = 0;

				Q931CallSendEvent(call, &event);

				Q931Log(pTrunk, Q931_LOG_DEBUG, "Garbage-collecting call %d\n", Q931CallGetCRV(call));

				/* Release call */
				Q931CallRelease(call);
			}
		}
	}

	Q931Log(pTrunk, Q931_LOG_DEBUG, "Q931Rx43 return code: %d\n", RetCode);
	return RetCode;
}

/*****************************************************************************

  Function:	 Q931Tx32

  Description:  Called from the stack to send a message to L2. The input is 
				always a non-packed message so it will first make a proper
				call to create a packed message before it transmits that
				message to layer 2.

  Parameters:   pTrunk[IN]  Trunk #  
				buf[IN]	 Ptr to message buffer.
				Size[IN]	Message size in bytes.

  Return Value: Error Code. 0 = No Error, < 0 :error, > 0 : Warning
				see q931errors.h for details.

*****************************************************************************/
L3INT Q931Tx32Data(Q931_TrunkInfo_t *pTrunk, L3UCHAR bcast, L3UCHAR * Mes, L3INT Size)
{
	Q931mes_Generic *ptr = (Q931mes_Generic*)Mes;
	L3INT RetCode = Q931E_NO_ERROR;
	L3INT Offset = bcast ? (Q931L2HeaderSpace - 1) : Q931L2HeaderSpace;
	L3INT OSize;
	L3UCHAR tei = 0;

	Q931Log(pTrunk, Q931_LOG_DEBUG, "Sending message to Q.921 (size: %d)\n", Size);

	memset(pTrunk->L2Buf, 0, sizeof(pTrunk->L2Buf));

	if (ptr->CRV) {
		struct Q931_Call *call;

		/* Find the call using CRV */
		call = Q931GetCallByCRV(pTrunk, ptr->CRV);
		if(!call)
			return Q931E_INVALID_CRV;

		tei = call->Tei;

		/* SAFEGUARD: make sure the CRVFlag is set correctly */
		ptr->CRVFlag = Q931CallIsOutgoing(call) ? 0 : 1;
	}

	/* Call pack function through table. */
	RetCode = Q931Pmes(pTrunk, ptr->MesType, (Q931mes_Generic *)Mes, Size, &pTrunk->L2Buf[Offset], &OSize);
	if (RetCode >= Q931E_NO_ERROR) {

		if (pTrunk->Q931Tx32CBProc) {
			RetCode = pTrunk->Q931Tx32CBProc(pTrunk->PrivateData32, bcast ? Q921_DL_UNIT_DATA : Q921_DL_DATA, tei, pTrunk->L2Buf, OSize + Offset);
			if (RetCode > 0) {
				/* L2 returns number of packets sent */
				RetCode = Q931E_NO_ERROR;
			}
		} else {
			RetCode = Q931E_MISSING_CB;
		}
	}

	return RetCode;
}


/*****************************************************************************

  Function:	 Q931SetError

  Description:  Called from the stack to indicate an error. 

  Parameters:   ErrID	   ID of ie or message causing error.
		ErrPar1	 Error parameter 1
		ErrPar2	 Error parameter 2.


*****************************************************************************/
void Q931SetError(Q931_TrunkInfo_t *pTrunk,L3INT ErrID, L3INT ErrPar1, L3INT ErrPar2)
{
	if (pTrunk->Q931ErrorCBProc) {
		pTrunk->Q931ErrorCBProc(pTrunk->PrivateData34, ErrID, ErrPar1, ErrPar2);
	}
}


L3INT Q931ReleaseCRV(Q931_TrunkInfo_t *pTrunk, L3INT CRV)
{
	struct Q931_Call *call;

	call = Q931GetCallByCRV(pTrunk, CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	call->InUse = 0;
	return Q931E_NO_ERROR;
}

L3ULONG Q931GetTime()
{
	L3ULONG tNow = 0;
	static L3ULONG tLast = 0;

	if (Q931GetTimeProc != NULL) {
		tNow = Q931GetTimeProc();
		if (tNow < tLast) {	/* wrapped */
			/* TODO */
		}
		tLast = tNow;
	}
	return tNow;
}

Q931_API(void) Q931SetGetTimeCB(L3ULONG (*callback)(void))
{
	Q931GetTimeProc = callback;
}


/*****************************************************************************
  Function:	 Q931AddStateEntry

  Description:  Find an empty entry in the dialects state table and add this
				entry.
*****************************************************************************/
void Q931AddStateEntry(L3UCHAR iD, L3INT iState, L3INT iMes, L3UCHAR flags)
{
	return;
}

/*****************************************************************************
  Function:	 Q931IsEventLegal

  Description:  Check state table for matching criteria to indicate if this
				Message is legal in this state or not.

  Note:		 Someone write a bsearch or invent something smart here
				please - sequential is ok for now.
*****************************************************************************/
L3BOOL Q931IsEventLegal(L3UCHAR iD, L3INT iState, L3INT iMes, L3UCHAR dirflags)
{
	return L3FALSE;
}

/*****************************************************************************
  Function:	 q931_error_to_name()

  Description:  Check state table for matching criteria to indicate if this
				Message is legal in this state or not.

  Note:		 Someone write a bsearch or invent something smart here
				please - sequential is ok for now.
*****************************************************************************/
static const char *q931_error_names[] = {
	"Q931E_NO_ERROR",			/* 0 */

	"Q931E_UNKNOWN_MESSAGE",		/* -3001 */
	"Q931E_ILLEGAL_IE",			/* -3002 */
	"Q931E_UNKNOWN_IE",			/* -3003 */
	"Q931E_BEARERCAP",			/* -3004 */
	"Q931E_HLCOMP",				/* -3005 */
	"Q931E_LLCOMP",				/* -3006 */
	"Q931E_INTERNAL",			/* -3007 */
	"Q931E_MISSING_CB",			/* -3008 */
	"Q931E_UNEXPECTED_MESSAGE",		/* -3009 */
	"Q931E_ILLEGAL_MESSAGE",		/* -3010 */
	"Q931E_TOMANYCALLS",			/* -3011 */
	"Q931E_INVALID_CRV",			/* -3012 */
	"Q931E_CALLID",				/* -3013 */
	"Q931E_CALLSTATE",			/* -3014 */
	"Q931E_CALLEDSUB",			/* -3015 */
	"Q931E_CALLEDNUM",			/* -3016 */
	"Q931E_CALLINGNUM",			/* -3017 */
	"Q931E_CALLINGSUB",			/* -3018 */
	"Q931E_CAUSE",				/* -3019 */
	"Q931E_CHANID",				/* -3020 */
	"Q931E_DATETIME",			/* -3021 */
	"Q931E_DISPLAY",			/* -3022 */
	"Q931E_KEYPADFAC",			/* -3023 */
	"Q931E_NETFAC",				/* -3024 */
	"Q931E_NOTIFIND",			/* -3025 */
	"Q931E_PROGIND",			/* -3026 */
	"Q931E_RESTARTIND",			/* -3027 */
	"Q931E_SEGMENT",			/* -3028 */
	"Q931E_SIGNAL",				/* -3029 */
	"Q931E_GENERIC_DIGITS",			/* -3030 */

	"Q931E_MANDATORY_IE_MISSING",		/* -3031 */
	"Q931E_UNRECOGNIZED_MESSAGE",		/* -3032 */
};

#define Q931_MAX_ERROR 32

const char *q931_error_to_name(q931_error_t error)
{
	int index = 0;
	if ((int)error < 0) {
		index = (((int)error * -1) -3000);
	}
	if (index < 0 || index > Q931_MAX_ERROR) {
		return "";
	}
	return q931_error_names[index];
}
/*
 * Logging
 */
#include <stdarg.h>

L3INT Q931Log(Q931_TrunkInfo_t *trunk, Q931LogLevel_t level, const char *fmt, ...)
{
	char buf[Q931_LOGBUFSIZE];
	L3INT len;
	va_list ap;

	if (!trunk->Q931LogCBProc)
		return 0;

	if (trunk->loglevel < level)
		return 0;

	va_start(ap, fmt);

	len = vsnprintf(buf, sizeof(buf)-1, fmt, ap);
	if (len <= 0) {
		/* TODO: error handling */
		return -1;
	}
	if (len >= sizeof(buf))
		len = sizeof(buf) - 1;

	buf[len] = '\0';

	va_end(ap);

	return trunk->Q931LogCBProc(trunk->PrivateDataLog, level, buf, len);
}

/**
 * Q921SetLogCB
 * \brief	Set Logging callback function and private data
 */
Q931_API(void) Q931SetLogCB(Q931_TrunkInfo_t *trunk, Q931LogCB_t func, void *priv)
{
	trunk->Q931LogCBProc  = func;
	trunk->PrivateDataLog = priv;
}

/**
 * Q921SetLogLevel
 * \brief	Set Loglevel
 */
Q931_API(void) Q931SetLogLevel(Q931_TrunkInfo_t *trunk, Q931LogLevel_t level)
{
	if(!trunk) {
		return;
	}

	if (level < Q931_LOG_NONE) {
		level = Q931_LOG_NONE;
	} else if (level > Q931_LOG_DEBUG) {
		level = Q931_LOG_DEBUG;
	}
	trunk->loglevel = level;
}

/**
 * Q931TimeoutDummy
 * \brief	Dummy handler for timeouts
 * \param	pTrunk		Q.931 trunk
 * \param	call		Q.931 call
 */
L3INT Q931TimeoutDummy(Q931_TrunkInfo_t *pTrunk, struct Q931_Call *call)
{
	Q931Log(pTrunk, Q931_LOG_DEBUG, "Timer %d of call %d timed out\n", call->TimerID, call->CRV);

	return 0;
}



/**************************************************************************************
 *  Trunk helper functions
 **************************************************************************************/
#include <assert.h>

void Q931TrunkSetFlag(Q931_TrunkInfo_t *trunk, L3INT flag)
{
	trunk->flags |= flag;
}

void Q931TrunkClearFlag(Q931_TrunkInfo_t *trunk, L3INT flag)
{
	trunk->flags &= ~flag;
}

void Q931TrunkClearAllFlags(Q931_TrunkInfo_t *trunk)
{
	trunk->flags = 0;
}

L3BOOL Q931TrunkIsSetFlag(const Q931_TrunkInfo_t *trunk, const L3INT flag)
{
	return ((trunk->flags & flag) != 0);
}

/* High-level access */
#define Q931_TRUNK_FLAG_SET_FUNC(_name, _flag) 			\
	Q931_API(void) _name(Q931_TrunkInfo_t *trunk, L3BOOL enable) {	\
		assert(trunk);					\
		if (enable) {					\
			Q931TrunkSetFlag(trunk, _flag);		\
		} else {					\
			Q931TrunkClearFlag(trunk, _flag);	\
		}						\
	}

#define Q931_TRUNK_FLAG_GET_FUNC(_name, _flag) 			\
	Q931_API(L3BOOL) _name(Q931_TrunkInfo_t *trunk) {		 	\
		assert(trunk);					\
		return Q931TrunkIsSetFlag(trunk, _flag);	\
	}

Q931_TRUNK_FLAG_SET_FUNC(Q931TrunkSetAutoSetupAck,     Q931_TFLAG_AUTO_SETUP_ACK);
Q931_TRUNK_FLAG_SET_FUNC(Q931TrunkSetAutoServiceAck,   Q931_TFLAG_AUTO_SERVICE_ACK);
Q931_TRUNK_FLAG_SET_FUNC(Q931TrunkSetAutoConnectAck,   Q931_TFLAG_AUTO_CONNECT_ACK);
Q931_TRUNK_FLAG_SET_FUNC(Q931TrunkSetAutoRestartAck,   Q931_TFLAG_AUTO_RESTART_ACK);
Q931_TRUNK_FLAG_SET_FUNC(Q931TrunkSetStatusEnquiry,    Q931_TFLAG_STATUS_ENQUIRY);
Q931_TRUNK_FLAG_SET_FUNC(Q931TrunkSetIgnoreUnknownMsg, Q931_TFLAG_IGNORE_UNKNOWN_MSG);
Q931_TRUNK_FLAG_SET_FUNC(Q931TrunkSetIgnoreUnknownIEs, Q931_TFLAG_IGNORE_UNKNOWN_IE);
Q931_TRUNK_FLAG_SET_FUNC(Q931TrunkSetIgnoreIllegalIEs, Q931_TFLAG_IGNORE_ILLEGAL_IE);

Q931_TRUNK_FLAG_GET_FUNC(Q931TrunkGetAutoSetupAck,     Q931_TFLAG_AUTO_SETUP_ACK);
Q931_TRUNK_FLAG_GET_FUNC(Q931TrunkGetAutoServiceAck,   Q931_TFLAG_AUTO_SERVICE_ACK);
Q931_TRUNK_FLAG_GET_FUNC(Q931TrunkGetAutoConnectAck,   Q931_TFLAG_AUTO_CONNECT_ACK);
Q931_TRUNK_FLAG_GET_FUNC(Q931TrunkGetAutoRestartAck,   Q931_TFLAG_AUTO_RESTART_ACK);
Q931_TRUNK_FLAG_GET_FUNC(Q931TrunkGetStatusEnquiry,    Q931_TFLAG_STATUS_ENQUIRY);
Q931_TRUNK_FLAG_GET_FUNC(Q931TrunkGetIgnoreUnknownMsg, Q931_TFLAG_IGNORE_UNKNOWN_MSG);
Q931_TRUNK_FLAG_GET_FUNC(Q931TrunkGetIgnoreUnknownIEs, Q931_TFLAG_IGNORE_UNKNOWN_IE);
Q931_TRUNK_FLAG_GET_FUNC(Q931TrunkGetIgnoreIllegalIEs, Q931_TFLAG_IGNORE_ILLEGAL_IE);

/**
 * Q931TrunkSetRelaxedMode
 * \brief	Relax IE and Message checks: incoming unknown messages and unknown/invalid IEs will be ignored
 * \param	trunk	Q.931 Trunk handle
 * \param	enable	Enable or disable Relaxed mode (Q931_TRUE / Q931_FALSE)
 *
 * \note	Use with care!
 */
Q931_API(void) Q931TrunkSetRelaxedMode(Q931_TrunkInfo_t *trunk, L3BOOL enable)
{
	int flags = (Q931_TFLAG_IGNORE_UNKNOWN_MSG | Q931_TFLAG_IGNORE_UNKNOWN_IE | Q931_TFLAG_IGNORE_ILLEGAL_IE);
	assert(trunk);

	if (enable) {
		Q931TrunkSetFlag(trunk, flags);
	} else {
		Q931TrunkClearFlag(trunk, flags);
	}
}


/***********************************************************************
 * Message, IE, Timer callback helpers
 ***********************************************************************/
L3INT Q931Proc(Q931_TrunkInfo_t *pTrunk, L3UCHAR *buf, L3INT iFrom)
{
	Q931mes_Header  *pMes = (Q931mes_Header *)&buf[Q931L4HeaderSpace];
	q931proc_func_t *func = Q931DialectGetMesProc(pTrunk->Dialect, pMes->MesType);

	if (func) {
		return func(pTrunk, buf, iFrom);
	}

        /* Ignore unknown message? */
        if (Q931TrunkIsSetFlag(pTrunk, Q931_TFLAG_IGNORE_UNKNOWN_MSG)) {
		return Q931E_NO_ERROR;
        }

	func = Q931DialectGetUnknownMesProc(pTrunk->Dialect);
	if (func) {
		return func(pTrunk, buf, iFrom);
	}
	return Q931E_INTERNAL;
}

L3INT Q931ProcInvalid(Q931_TrunkInfo_t *pTrunk, L3UCHAR *buf, L3INT iFrom)
{
	q931proc_func_t *func = Q931DialectGetInvalidMesProc(pTrunk->Dialect);

	if (func) {
		return func(pTrunk, buf, iFrom);
	}
	return Q931E_INTERNAL;	
}

L3INT Q931Umes(Q931_TrunkInfo_t *pTrunk, L3UCHAR id, L3UCHAR *IBuf, Q931mes_Generic *OBuf, L3INT IOff, L3INT Size)
{
	q931umes_func_t *func = Q931DialectGetUmesProc(pTrunk->Dialect, id);

	if (func) {
		return func(pTrunk, IBuf, OBuf, IOff, Size);
	}

        /* Ignore unknown message? */
        if (Q931TrunkIsSetFlag(pTrunk, Q931_TFLAG_IGNORE_UNKNOWN_MSG)) {
		return Q931E_NO_ERROR;
        }

	return Q931E_INTERNAL;
}

L3INT Q931Pmes(Q931_TrunkInfo_t *pTrunk, L3UCHAR id, Q931mes_Generic *IBuf, L3INT ISize, L3UCHAR *OBuf, L3INT *OSize)
{
	q931pmes_func_t *func = Q931DialectGetPmesProc(pTrunk->Dialect, id);

	if (func) {
		return func(pTrunk, IBuf, ISize, OBuf, OSize);
	}
	return Q931E_INTERNAL;
}

L3INT Q931Uie(Q931_TrunkInfo_t *pTrunk, L3UCHAR id, Q931mes_Generic *pMes, L3UCHAR *IBuf, L3UCHAR *OBuf, L3INT *IOff, L3INT *OOff)
{
	q931uie_func_t *func = Q931DialectGetUieProc(pTrunk->Dialect, id);

	if (func) {
		return func(pTrunk, pMes, IBuf, OBuf, IOff, OOff);
	}
	return Q931E_INTERNAL;
}

L3INT Q931Pie(Q931_TrunkInfo_t *pTrunk, L3UCHAR id, L3UCHAR *IBuf, L3UCHAR *OBuf, L3INT *Octet)
{
	q931pie_func_t *func = Q931DialectGetPieProc(pTrunk->Dialect, id);

	if (func) {
		return func(pTrunk, IBuf, OBuf, Octet);
	}
	return Q931E_INTERNAL;
}

L3BOOL Q931UieIsNull(Q931_TrunkInfo_t *pTrunk, L3UCHAR id)
{
	return (!Q931DialectGetUieProc(pTrunk->Dialect, id));
}

L3BOOL Q931PieIsNull(Q931_TrunkInfo_t *pTrunk, L3UCHAR id)
{
	return (!Q931DialectGetPieProc(pTrunk->Dialect, id));
}

L3INT Q931Timeout(Q931_TrunkInfo_t *pTrunk, struct Q931_Call *call, L3UCHAR timer)
{
	q931timeout_func_t *func = Q931DialectGetTimerProc(pTrunk->Dialect, timer);

	if (func) {
		return func(pTrunk, call);
	}
	return Q931E_INTERNAL;
}


/*
 * Default handler for unexpected / unknown messages
 */
L3INT Q931ProcUnknownMessage(Q931_TrunkInfo_t *pTrunk, L3UCHAR *buf, L3INT iFrom)
{
	Q931mes_Header *mes = (Q931mes_Header *)&buf[Q931L4HeaderSpace];

	Q931Log(pTrunk, Q931_LOG_NOTICE, "Unknown message %#hx received\n", mes->MesType);

	return Q931E_NO_ERROR;
}

L3INT Q931ProcUnexpectedMessage(Q931_TrunkInfo_t *pTrunk, L3UCHAR *buf, L3INT iFrom)
{
	Q931mes_Header *mes = (Q931mes_Header *)&buf[Q931L4HeaderSpace];

	Q931Log(pTrunk, Q931_LOG_NOTICE, "Unexpected/Invalid message %#hx received\n", mes->MesType);

	return Q931E_NO_ERROR;
}

